// qdtexture.cpp
// by Glenn G. Chappell
// December 2003
//
// For CS 381
// Quick & dirty texture generation from a string array

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff, includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// For texture
const int img_width = 8;
const int img_height = 8;
GLubyte the_image[img_height][img_width][3];
   // The image
   // 3rd subscript 0 = R, 1 = G, 2 = B
GLuint texname;  // Name for this texture
char text_image[img_height][img_width+1] = {  // width+1 for ending '\0'
   "--------",
   "--XXXX--",
   "-X----X-",
   "-X------",
   "-X--XXX-",
   "-X----X-",
   "--XXXXX-",
   "--------" };

// For user-controlled stuff
bool textureon = true; // True if we are texturing

double zoom = 1.0;    // scale factor
const double minzoom = 0.1;
const double maxzoom = 10.0;
const double zoomfactor = 1.1; // Mult. factor for zoom increase

bool rotate = false;  // Is rotating?
double angle = 0.0;   // Current rotation angle (degrees)
const double anglestep = 1.0;  // Step for above (degrees)


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT);

   // Draw square
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   glScaled(5, 5, 5);
   glMatrixMode(GL_MODELVIEW);

   glPushMatrix();

      glRotated(angle, 0,0,1);
      glScaled(zoom, zoom, zoom);

      if (textureon)
         glEnable(GL_TEXTURE_2D);
      else
         glDisable(GL_TEXTURE_2D);

      glColor3d(0.6, 0.4, 0.2);  // Not used in REPLACE-style texturing.


      glBegin(GL_POLYGON);
         glTexCoord2d(1.0, 0.0); glVertex2d( 0.7,-0.7);
         glTexCoord2d(1.0, 1.0); glVertex2d( 0.7, 0.7);
         glTexCoord2d(0.0, 1.0); glVertex2d(-0.7, 0.7);
         glTexCoord2d(0.0, 0.0); glVertex2d(-0.7,-0.7);
      glEnd();

      glDisable(GL_TEXTURE_2D);  // Don't want to affect colors of remaining stuff

   glPopMatrix();

   // Draw instructions
   glColor3d(0.0, 0.0, 0.0);
   printbitmap("T      Toggle texture", -0.9, 0.9);
   printbitmap("R      Toggle rotation", -0.9, 0.8);
   printbitmap("<- ->  Zoom", -0.9, 0.7);
   printbitmap("Esc    Quit", -0.9, 0.6);

   glutSwapBuffers();
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // We set up coordinate system so that aspect ratios are always correct,
    //  and the region from -1..1 in x & y always just fits in the viewport.
    if (w > h)
    {
       gluOrtho2D(-double(w)/h, double(w)/h, -1., 1.);
    }
    else
    {
       gluOrtho2D( -1., 1., -double(h)/w, double(h)/w);
    }

    glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 'r':
   case 'R':     // R: toggle rotation
      rotate = !rotate;
      glutPostRedisplay();
      break;
   case 't':
   case 'T':     // T: toggle texture
     textureon = !textureon;
     glutPostRedisplay();
     break;
   }
}


// special
// The GLUT special function
void special(int key, int x, int y)
{
   switch(key)
   {
   case GLUT_KEY_RIGHT:  // -> : Zoom goes up
      zoom *= zoomfactor;
      if (zoom > maxzoom) zoom = maxzoom;
      glutPostRedisplay();
      break;
   case GLUT_KEY_LEFT:   // <- : Zoom goes down
      zoom /= zoomfactor;
      if (zoom < minzoom) zoom = minzoom;
      glutPostRedisplay();
      break;
   }
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Do rotation
   if (rotate)
   {
      angle += anglestep;
      glutPostRedisplay();
   }
}


// makeimage
// Make image in the_image
//  using data in text_image
void makeimage()
{
   for (int i=0; i<img_width; ++i)
   {
      for (int j=0; j<img_height; ++j)
      {
         switch(text_image[j][i])
         {
         case 'X':
            // OpenGL stores images bottom-up,
            // so we "reverse" j:
            the_image[img_height-1-j][i][0] = 128;
            the_image[img_height-1-j][i][1] = 255;
            the_image[img_height-1-j][i][2] = 255;
            break;
         default:
            the_image[img_height-1-j][i][0] = 128;
            the_image[img_height-1-j][i][1] = 128;
            the_image[img_height-1-j][i][2] = 128;
            break;
         }
      }
   }
}


// init
// Initialization
// Called by main after window creation
void init()
{
   glClearColor(1.0, 1.0, 1.0, 1.0);

   // Texture set-up

   // Basic image set-up
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   // Generate texture name(s)
   glGenTextures(1, &texname);
   glBindTexture(GL_TEXTURE_2D, texname);

   // Set-up the texture
   makeimage();
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img_width, img_height,
                0, GL_RGB, GL_UNSIGNED_BYTE, the_image);

   // Set global texture states
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Quick & Dirty Texture Generation");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutSpecialFunc(special);
   glutIdleFunc(idle);

   // Do something
   glutMainLoop();

   return 0;
}
